grammar uppaal;

/*
              Uppaal Timed Automata --- Partial Grammar

  (I chose to rely on ANTLR instead of the parser combinators as when working
   with ANTLRWorks it is much simpler to get the grammar right...)
*/

options {	
 // language = Scala;
  k = 1;
}

@lexer::header
{
 /**
  *
  *                  Incremental TCG implementation
  *
  * Copyright 2012, Willibald Krenn, AIT Austrian Institute of Technology
  *
  * Do not modify this file. It was autogenerated by ANTLR.
  */
  package incrementalTCG.compiler.parser;  
}

@parser::header
{
 /**
  *
  *                  Incremental TCG implementation
  *
  * Copyright 2012, Willibald Krenn, AIT Austrian Institute of Technology
  *
  * Do not modify this file. It was autogenerated by ANTLR.
  */
package incrementalTCG.compiler.parser;  

import incrementalTCG.ast.*;
import java.util.LinkedList;
}

@parser::members
{
 /* Begin Custom Members */

	public final UppaalParserHelper helper = new UppaalParserHelper(); 

	@Override
	public void displayRecognitionError(String[] tokenNames, RecognitionException e) {
        	final String hdr = getErrorHeader(e);
        	final String msg = getErrorMessage(e, tokenNames);
		helper.addError(hdr, msg);
        }
 /* End Custom Members */
}

/* Special Case: Guard Declarations: */

guardDeclarations
	:	guardDeclaration (T_COMMA guardDeclaration)*
	;
guardDeclaration
	:	varid=T_IDENTIFIER T_COLON t=simpletype {helper.addVariable(varid,t,null,null,null);}
	;

/*Normal language:*/

declarations
	:	declaration*
	;

declaration 
	:	constantDeclaration // also deals with const vars..
	|	typeDeclaration     // typdef
	|	channelDeclaration  // chan
	|	variableOrFunctionDeclaration // everything else
	;

variableOrFunctionDeclaration
	:	t=simpletype varid=T_IDENTIFIER 		
		(
			{helper.beginParseFunction(varid,t);}
			T_LPAREN argumentList T_RPAREN functionBody
		|	T_LSQPAREN e=expression T_RSQPAREN T_SEMICOLON {helper.addVariable(varid,t,e,null,null);}
		|       (T_EQUAL expr=expression)? T_SEMICOLON {helper.addVariable(varid,t,null,null,expr);} 
		)
	;
	finally
		{helper.endParseFunction();}	

argumentList
	: ( ctok=T_CONST? tpe=simpletype nme=T_IDENTIFIER {helper.addArgument(nme,tpe,null,ctok);} 
	    (T_COMMA ctok2=T_CONST? tpe2=simpletype nme2=T_IDENTIFIER {helper.addArgument(nme2,tpe2,null,ctok2);})*)?
	;
	
functionBody
	: T_CBRL s1=statement {helper.addStatement(s1);} T_SEMICOLON (s2=statement T_SEMICOLON {helper.addStatement(s2);})* T_CBRR	
	;
	
statement
	returns [Expression statement]
	: e1=expression (a=assignoperator  e2=expression)?
	  {statement = helper.statement(e1,a,e2);}	  
	;

channelDeclaration
	:	T_CHAN chan_a=T_IDENTIFIER (T_LSQPAREN e=expression T_RSQPAREN)? 
			{helper.addChannel(chan_a, e);}
		
		(T_COMMA chan_b=T_IDENTIFIER (T_LSQPAREN e2=expression T_RSQPAREN)?
			{helper.addChannel(chan_b, e2);})* 
		
		T_SEMICOLON
	;



typeDeclaration
	:	T_TYPEDEF t=simpletype id=T_IDENTIFIER T_SEMICOLON {helper.addNamedType(id,t);}
	;


constantDeclaration
	:	T_CONST t=simpletype id=T_IDENTIFIER (T_EQUAL expr=expression)? T_SEMICOLON {helper.addNamedConstant(id,t,expr);}
	;


simpletype	
	returns [SimpleType type]
	:	T_INT   (T_LSQPAREN rangeLow=expression T_COMMA rangeHigh=expression T_RSQPAREN)?   {type = helper.intType(rangeLow, rangeHigh);}   // integer
	|	T_CLOCK         {type = helper.clockType();} // clock
	|	T_VOID		{type = helper.voidType();}
	|	id=T_IDENTIFIER {type = helper.opaqueType(id);} // alias
	;


//
//  ------------   expression    ------------ 
//
expression
	returns [Expression expr]
	@init{
		LinkedList<Token> operators = new LinkedList<Token>();
		LinkedList<Expression> expressions = new LinkedList<Expression>();		
	}
	:	e=atomExpression      {expressions.add(e);}
		( binop=binoperator
		  {operators.add(binop);}
		  e2=atomExpression
		  {expressions.add(e2);}
		)*
		{expr = helper.createPrecedenceTree(expressions,operators);}
	;

atomExpression
	returns [Expression expr]
	:	unop=op_un_pre? (   // FIXME!!!!
		     e=identifierExpression             {expr = e;}
		  |  e=constant                         {expr = e;}
		  |  T_LPAREN e=expression T_RPAREN     {expr = e;} 
		)
		{expr = helper.unop(unop,expr);}
	;	

identifierExpression
	returns [IdentifierExpression expr]
	:	id=T_IDENTIFIER {expr = helper.identifierExpression(id);}
		(
			T_LPAREN (e1=expression  {helper.addFunctionCallArg(expr, e1);} (T_COMMA e2=expression {helper.addFunctionCallArg(expr, e2);})*)? T_RPAREN
		|	T_LSQPAREN e3=expression T_RSQPAREN 
		        {helper.addArrayAccess(expr, e3);}
		)?
		( '?' {helper.inputSync(expr);} | '!' {helper.outputSync(expr);})?
	;


constant
	returns [Expression expr]
	: 	T_TRUE            {expr = helper.constantTrue();}
	|	T_FALSE           {expr = helper.constantFalse();}
	| 	n=T_FLOATNUMBER   {expr = helper.constantFloat(n);}
	|	i=T_INTNUMBER     {expr = helper.constantInteger(i);}
	;	

binoperator	
	returns [Token op]
	: o=T_SMALLEREQ {op = o;}
	| o=T_SMALLER   {op = o;}
	| o=T_EQ        {op = o;}
	| o=T_GREATEREQ {op = o;}
	| o=T_GREATER   {op = o;}
	| o=T_NEQ       {op = o;}
	| o=T_DAMP      {op = o;}
	| o=T_AMP       {op = o;}
	| o=T_BAR       {op = o;}
	| o=T_OR        {op = o;}
	| o=T_AND       {op = o;}
	| o=T_PLUS      {op = o;}
	| o=T_MINUS     {op = o;}
	| o=T_MUL       {op = o;}
	| o=T_DIV       {op = o;}
	| o=T_MOD       {op = o;}
	;
	

assignoperator
	returns [Token op]
	: o=T_EQUAL  {op = o;}
	| o=T_ASS    {op = o;}
	| o=T_PSASS  {op = o;}
	| o=T_MASS   {op = o;}
	| o=T_MULASS {op = o;}
	| o=T_DIVASS {op = o;}
	| o=T_MODASS {op = o;}
	| o=T_ORASS  {op = o;}
	| o=T_ANDASS {op = o;}
	| o=T_XORASS {op = o;}
	| o=T_LSHASS {op = o;}
	| o=T_RSHASS {op = o;}
	;


//
//  ------------  unary  Operators    ------------ 
//
	
	
op_un_pre
	returns [Token op]
	:	o=T_MINUS {op=o;} 	
	|	o=T_PLUS  {op=o;}
	|	o=T_NOT   {op=o;}		
	|	o=T_NEG   {op=o;}		
	;	




//
//  ==============   LEXER   =================
//


T_WS  	: (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=BaseRecognizer.HIDDEN;}
    	;
    	
T_COMMENT
    	: '/*' .* '*/' {$channel=BaseRecognizer.HIDDEN;}
    	;
    	
LINE_COMMENT
    	: '//' ~('\n'|'\r')* '\r'? '\n' {$channel=BaseRecognizer.HIDDEN;}
    	;

T_STRINGLITERAL
	:	'"' 
		(
     			( '"' '"' )=>  '"' 
      		|  	~'"'
  		)*
 		'"'
	;


T_ASS	: ':=' ;
T_PSASS	: '+=' ;
T_MASS	: '-=' ;
T_MULASS: '*=' ;
T_DIVASS: '/=' ;
T_MODASS: '%=' ;
T_ORASS	: '|=' ;
T_ANDASS: '&=' ;
T_XORASS: '^=' ;
T_LSHASS: '<<=';
T_RSHASS: '>>=';


T_AND	:	'and';
T_MUL	:	'*';
T_DIV	:	'/';
T_MOD	:	'%';
T_OR	:	'or';
T_DAMP	:	'&&';
T_AMP	:	'&';
T_NEQ	:	'!=';
T_GREATEREQ
	:	'>=';
T_GREATER
	:	'>';
T_SMALLEREQ
	:	'<=';
T_SMALLER
	:	'<';
T_EQ	:	'==';

T_BAR	:	'|';
T_CBRL	:	'{';
T_CBRR	:	'}';
T_COLON	:	':';
T_COMMA	:	',';
T_EQUAL	:	'=';
T_LPAREN:	'(';
T_RPAREN:	')';
T_LSQPAREN:	'[';
T_RSQPAREN:	']';
T_SEMICOLON
	:	';';


T_INT	:	'int';
T_CLOCK	:	'clock';
T_VOID	:	'void';


T_NOT	:	'not';
T_NEG	:	'!';
T_CONST	:	'const';
T_CHAN	:	'chan';
T_TYPEDEF
	:	'typedef';
T_TRUE	:	'true';
T_FALSE	:	'false';





// numbers

// INTNUMBER, FLOATNUMBER, RANGETO, T_MINUS, T_PLUS are set within
// FLOAT_OR_INT_OR_RANGE
fragment T_INTNUMBER
	:	;
fragment T_FLOATNUMBER
	:	;


FLOAT_OR_INT_OR_RANGE
	: 	T_DIGIT+
		( 		(T_POINT T_DIGIT) => T_POINT T_DIGIT+ (('e'|'E') (T_MINUS|T_PLUS)? T_DIGIT+)? {$type=T_FLOATNUMBER;}				
			|	{$type=T_INTNUMBER;}
		)
	|	T_POINT	{$type=T_POINT;}		
	|	T_MINUS	{$type=T_MINUS;}
	|	T_PLUS	{$type=T_PLUS;}
	;	

fragment T_MINUS :	'-';
fragment T_PLUS	:	'+';
fragment T_POINT:	'.';



// identifiers
T_IDENTIFIER
	:	T_LETTER (T_LETTER|T_DIGIT)* 
	;
	
//T_PRIMEDIDENTIFIER
//	:	T_IDENTIFIER '\'';
	
fragment
T_LETTER
	:	'$'
	|	'A'..'Z'
	|	'a'..'z'
	|	'_'
	;

fragment 
T_DIGIT:	'0'..'9';

